import time

import lal
import numpy as np
from astropy.time import Time
from bilby_cython.time import (
    greenwich_sidereal_time,
    greenwich_mean_sidereal_time,
    gps_time_to_utc,
    utc_to_julian_day,
)


def test_gps_time_to_julian_day(types):
    times = np.random.uniform(1325623903, 1345623903, 100000).astype(types)
    diffs = list()
    for tt in times:
        diffs.append(
            utc_to_julian_day(gps_time_to_utc(tt))
            - lal.JulianDay(lal.GPSToUTC(int(tt)))
        )
    assert max(np.abs(diffs)) < 1e-10


def test_gmst(types):
    times = np.random.uniform(1325623903, 1345623903, 100000).astype(types)
    diffs = list()
    for tt in times:
        diffs.append(
            greenwich_mean_sidereal_time(tt)
            - lal.GreenwichMeanSiderealTime(tt)
        )
    assert max(np.abs(diffs)) < 1e-10


def test_gmst_vectorized(types):
    times = np.random.uniform(1325623903, 1345623903, 100000).astype(types)
    diffs = list()
    cy_gmst = greenwich_mean_sidereal_time(times)
    lal_gmst = np.array([lal.GreenwichMeanSiderealTime(tt) for tt in times])
    assert max(np.abs(cy_gmst - lal_gmst)) < 1e-10


def test_gmt(types):
    times = np.random.uniform(1325623903, 1345623903, 100000).astype(types)
    equinoxes = np.random.uniform(0, 2 * np.pi, 100000)
    diffs = list()
    for tt, eq in zip(times, equinoxes):
        diffs.append(
            greenwich_sidereal_time(tt, eq)
            - lal.GreenwichSiderealTime(tt, eq)
        )
    assert max(np.abs(diffs)) < 1e-10


def test_current_time():
    """
    Test that the current GMST matches LAL and Astropy.
    This should ensure robustness against additional leap seconds being added. 
    """
    now = float(lal.GPSTimeNow())
    lal_now = lal.GreenwichMeanSiderealTime(now) % (2 * np.pi)
    cython_now = greenwich_mean_sidereal_time(now) % (2 * np.pi)
    astropy_now = Time(now, format="gps").sidereal_time("mean", 0.0).radian
    assert np.abs(cython_now - lal_now) < 1e-10
    assert np.abs(cython_now - astropy_now) < 1e-5



import time

import numpy as np

from bilby.gw.detector import InterferometerList
from bilby_cython import geometry
from bilby_cython.time import greenwich_mean_sidereal_time
from old_code import (
    antenna_response,
    get_polarization_tensor,
    get_polarization_tensor_multiple_modes,
    greenwich_mean_sidereal_time as gmst_old,
    time_delay_geocentric,
    zenith_azimuth_to_theta_phi,
)

ifos = InterferometerList(["H1", "L1"])
detectors = [ifo.vertex for ifo in ifos]

points = np.random.uniform(0, np.pi / 2, (10000, 3))

start = time.time()
for point in points:
    time_delay_geocentric(*detectors, *point)
stop = time.time()
numpy_time = (stop - start) / len(points)
start = time.time()
for point in points:
    geometry.time_delay_geocentric(*detectors, *point)
stop = time.time()
cython_time = (stop - start) / len(points)

print(f"Timing time delay calculation over {len(points)} trials.")
print(f"Cython time: {cython_time:.3e}\nNumpy time: {numpy_time:.3e}")

points = np.random.uniform(0, np.pi / 2, (1000, 4))
modes = ["plus", "cross", "x", "y", "breathing", "longitudinal"]

start = time.time()
for point in points:
    for mode in modes:
        get_polarization_tensor(*point, mode)
stop = time.time()
numpy_time = (stop - start) / len(points) / len(modes)
start = time.time()
for point in points:
    for mode in modes:
        geometry.get_polarization_tensor(*point, mode)
stop = time.time()
cython_time = (stop - start) / len(points) / len(modes)

print(f"Timing polarization tensor calculation over {len(points)} trials.")
print(f"Cython time: {cython_time:.3e}\nNumpy time: {numpy_time:.3e}")

points = np.random.uniform(0, np.pi / 2, (1000, 4))
modes = ["plus", "cross"]

start = time.time()
for point in points:
    get_polarization_tensor_multiple_modes(*point, modes)
stop = time.time()
numpy_time = (stop - start) / len(points) / len(modes)
start = time.time()
for point in points:
    geometry.get_polarization_tensor_multiple_modes(*point, modes)
stop = time.time()
cython_time = (stop - start) / len(points) / len(modes)

print(f"Timing two mode polarization tensor calculation over {len(points)} trials.")
print(f"Cython time: {cython_time:.3e}\nNumpy time: {numpy_time:.3e}")


points = np.random.uniform(0, np.pi / 2, (1000, 4))
modes = ["plus", "cross", "x", "y", "breathing", "longitudinal"]

start = time.time()
for point in points:
    get_polarization_tensor_multiple_modes(*point, modes)
stop = time.time()
numpy_time = (stop - start) / len(points) / len(modes)
start = time.time()
for point in points:
    geometry.get_polarization_tensor_multiple_modes(*point, modes)
stop = time.time()
cython_time = (stop - start) / len(points) / len(modes)

print(f"Timing six mode polarization tensor calculation over {len(points)} trials.")
print(f"Cython time: {cython_time:.3e}\nNumpy time: {numpy_time:.3e}")


def cython_antenna_response(detector_tensor, ra, dec, time, psi, mode):
    polarization = geometry.get_polarization_tensor(ra, dec, time, psi, mode)
    return geometry.three_by_three_matrix_contraction(detector_tensor, polarization)


points = np.random.uniform(0, np.pi / 2, (1000, 4))
modes = ["plus", "cross", "x", "y", "breathing", "longitudinal"]

start = time.time()
for point in points:
    for mode in modes:
        antenna_response(ifos[0].geometry.detector_tensor, *point, mode)
stop = time.time()
numpy_time = (stop - start) / len(points) / len(modes)
start = time.time()
for point in points:
    for mode in modes:
        cython_antenna_response(ifos[0].geometry.detector_tensor, *point, mode)
stop = time.time()
cython_time = (stop - start) / len(points) / len(modes)

print(f"Timing antenna response calculation over {len(points)} trials.")
print(f"Cython time: {cython_time:.3e}\nNumpy time: {numpy_time:.3e}")

points = np.random.uniform(0, np.pi, (1000, 2))
delta_x = ifos[1].geometry.vertex - ifos[0].geometry.vertex
# burn a numpy evaluation for caching
zenith_azimuth_to_theta_phi(*points[0], delta_x)

start = time.time()
for point in points:
    zenith_azimuth_to_theta_phi(*point, delta_x)
stop = time.time()
numpy_time = (stop - start) / len(points)
start = time.time()
for point in points:
    geometry.zenith_azimuth_to_theta_phi(*point, delta_x)
stop = time.time()
cython_time = (stop - start) / len(points)

print(f"Timing frame conversion calculation over {len(points)} trials.")
print(f"Cython time: {cython_time:.3e}\nNumpy time: {numpy_time:.3e}")

times = np.random.uniform(1000000000, 1000100000, 10000)
start = time.time()
for tt in times:
    gmst_old(tt)
stop = time.time()
lal_time = (stop - start) / len(times)
start = time.time()
for tt in times:
    greenwich_mean_sidereal_time(tt)
stop = time.time()
cython_time = (stop - start) / len(times)

print(f"Timing GMST calculation over {len(points)} trials.")
print(f"Cython time: {cython_time:.3e}\nLAL time: {lal_time:.3e}")



from math import fmod

import numpy as np


def calculate_arm(arm_tilt, arm_azimuth, longitude, latitude):
    """
    Pulled from bilby v1.1.5
    """
    e_long = np.array([-np.sin(longitude), np.cos(longitude), 0])
    e_lat = np.array([-np.sin(latitude) * np.cos(longitude),
                      -np.sin(latitude) * np.sin(longitude), np.cos(latitude)])
    e_h = np.array([np.cos(latitude) * np.cos(longitude),
                    np.cos(latitude) * np.sin(longitude), np.sin(latitude)])

    return (np.cos(arm_tilt) * np.cos(arm_azimuth) * e_long +
            np.cos(arm_tilt) * np.sin(arm_azimuth) * e_lat +
            np.sin(arm_tilt) * e_h)


def time_delay_geocentric(detector1, detector2, ra, dec, time):
    """
    Copied from Bilby v1.1.4.

    Calculate time delay between two detectors in geocentric coordinates based on XLALArrivaTimeDiff in TimeDelay.c

    Parameters
    ==========
    detector1: array_like
        Cartesian coordinate vector for the first detector in the geocentric frame
        generated by the Interferometer class as self.vertex.
    detector2: array_like
        Cartesian coordinate vector for the second detector in the geocentric frame.
        To get time delay from Earth center, use detector2 = np.array([0,0,0])
    ra: float
        Right ascension of the source in radians
    dec: float
        Declination of the source in radians
    time: float
        GPS time in the geocentric frame

    Returns
    =======
    float: Time delay between the two detectors in the geocentric frame

    """
    from lal import GreenwichMeanSiderealTime
    speed_of_light = 299792458.0
    gmst = fmod(GreenwichMeanSiderealTime(time), 2 * np.pi)
    phi = ra - gmst
    theta = np.pi / 2 - dec
    omega = np.array([np.sin(theta) * np.cos(phi), np.sin(theta) * np.sin(phi), np.cos(theta)])
    delta_d = detector2 - detector1
    return np.dot(omega, delta_d) / speed_of_light


def get_polarization_tensor(ra, dec, time, psi, mode):
    """
    Copied from Bilby v1.1.4.

    Calculate the polarization tensor for a given sky location and time

    See Nishizawa et al. (2009) arXiv:0903.0528 for definitions of the polarisation tensors.
    [u, v, w] represent the Earth-frame
    [m, n, omega] represent the wave-frame
    Note: there is a typo in the definition of the wave-frame in Nishizawa et al.

    Parameters
    ==========
    ra: float
        right ascension in radians
    dec: float
        declination in radians
    time: float
        geocentric GPS time
    psi: float
        binary polarisation angle counter-clockwise about the direction of propagation
    mode: str
        polarisation mode

    Returns
    =======
    array_like: A 3x3 representation of the polarization_tensor for the specified mode.

    """
    from lal import GreenwichMeanSiderealTime
    gmst = fmod(GreenwichMeanSiderealTime(time), 2 * np.pi)
    phi = ra - gmst
    theta = np.pi / 2 - dec
    u = np.array([np.cos(phi) * np.cos(theta), np.cos(theta) * np.sin(phi), -np.sin(theta)])
    v = np.array([-np.sin(phi), np.cos(phi), 0])
    m = -u * np.sin(psi) - v * np.cos(psi)
    n = -u * np.cos(psi) + v * np.sin(psi)

    if mode.lower() == 'plus':
        return np.einsum('i,j->ij', m, m) - np.einsum('i,j->ij', n, n)
    elif mode.lower() == 'cross':
        return np.einsum('i,j->ij', m, n) + np.einsum('i,j->ij', n, m)
    elif mode.lower() == 'breathing':
        return np.einsum('i,j->ij', m, m) + np.einsum('i,j->ij', n, n)

    # Calculating omega here to avoid calculation when model in [plus, cross, breathing]
    omega = np.cross(m, n)
    if mode.lower() == 'longitudinal':
        return np.einsum('i,j->ij', omega, omega)
    elif mode.lower() == 'x':
        return np.einsum('i,j->ij', m, omega) + np.einsum('i,j->ij', omega, m)
    elif mode.lower() == 'y':
        return np.einsum('i,j->ij', n, omega) + np.einsum('i,j->ij', omega, n)
    else:
        raise ValueError("{} not a polarization mode!".format(mode))


def get_polarization_tensor_multiple_modes(ra, dec, time, psi, modes):
    return [get_polarization_tensor(ra, dec, time, psi, mode) for mode in modes]


def antenna_response(detector_tensor, ra, dec, time, psi, mode):
    polarization_tensor = get_polarization_tensor(ra, dec, time, psi, mode)
    return np.einsum('ij,ij->', detector_tensor, polarization_tensor)


__cached_euler_matrix = None
__cached_delta_x = None


def euler_rotation(delta_x):
    """
    Calculate the rotation matrix mapping the vector (0, 0, 1) to delta_x
    while preserving the origin of the azimuthal angle.

    This is decomposed into three Euler angle, alpha, beta, gamma, which rotate
    about the z-, y-, and z- axes respectively.

    Parameters
    ==========
    delta_x: array-like (3,)
        Vector onto which (0, 0, 1) should be mapped.

    Returns
    =======
    total_rotation: array-like (3,3)
        Rotation matrix which maps vectors from the frame in which delta_x is
        aligned with the z-axis to the target frame.
    """
    global __cached_delta_x
    global __cached_euler_matrix

    delta_x = delta_x / np.sum(delta_x**2)**0.5
    if np.array_equal(delta_x, __cached_delta_x):
        return __cached_euler_matrix
    else:
        __cached_delta_x = delta_x
    alpha = np.arctan(- delta_x[1] * delta_x[2] / delta_x[0])
    beta = np.arccos(delta_x[2])
    gamma = np.arctan(delta_x[1] / delta_x[0])
    rotation_1 = np.array([
        [np.cos(alpha), -np.sin(alpha), 0], [np.sin(alpha), np.cos(alpha), 0],
        [0, 0, 1]])
    rotation_2 = np.array([
        [np.cos(beta), 0, - np.sin(beta)], [0, 1, 0],
        [np.sin(beta), 0, np.cos(beta)]])
    rotation_3 = np.array([
        [np.cos(gamma), -np.sin(gamma), 0], [np.sin(gamma), np.cos(gamma), 0],
        [0, 0, 1]])
    total_rotation = np.einsum(
        'ij,jk,kl->il', rotation_3, rotation_2, rotation_1)
    __cached_delta_x = delta_x
    __cached_euler_matrix = total_rotation
    return total_rotation


def zenith_azimuth_to_theta_phi(zenith, azimuth, delta_x):
    """
    Convert from the 'detector frame' to the Earth frame.

    Parameters
    ==========
    zenith: float
        The zenith angle in the detector frame
    azimuth: float
        The azimuthal angle in the detector frame
    delta_x: list
        List of Interferometer objects defining the detector frame

    Returns
    =======
    theta, phi: float
        The zenith and azimuthal angles in the earth frame.
    """
    omega_prime = np.array([
        np.sin(zenith) * np.cos(azimuth),
        np.sin(zenith) * np.sin(azimuth),
        np.cos(zenith)])
    rotation_matrix = euler_rotation(delta_x)
    omega = np.dot(rotation_matrix, omega_prime)
    theta = np.arccos(omega[2])
    phi = np.arctan2(omega[1], omega[0]) % (2 * np.pi)
    return theta, phi


def greenwich_mean_sidereal_time(time):
    from lal import GreenwichMeanSiderealTime
    time = float(time)
    return GreenwichMeanSiderealTime(time)



import itertools
import unittest

import numpy as np
from bilby.gw.detector import get_empty_interferometer, InterferometerList
from bilby_cython import geometry

from .old_code import (
    antenna_response,
    calculate_arm,
    get_polarization_tensor,
    get_polarization_tensor_multiple_modes,
    time_delay_geocentric,
    zenith_azimuth_to_theta_phi,
)


class TestGeometry(unittest.TestCase):

    def setUp(self) -> None:
        self.ifos = ["H1", "L1", "V1", "K1"]
        self.modes = ["plus", "cross", "x", "y", "breathing", "longitudinal"]

    def tearDown(self) -> None:
        pass

    def test_time_delay(self):
        max_diff = 0
        for ifo_pair in itertools.product(self.ifos, repeat=2):
            if ifo_pair[0] == ifo_pair[1]:
                continue
            ifos = InterferometerList(ifo_pair)
            detectors = [ifo.vertex for ifo in ifos]
            for point in np.random.uniform(0, np.pi / 2, (1000, 3)):
                numpy_delay = time_delay_geocentric(*detectors, *point)
                cython_delay = geometry.time_delay_geocentric(*detectors, *point)
                max_diff = max(max_diff, abs(numpy_delay - cython_delay))
        self.assertLess(max_diff, 1e-6)

    def test_time_delay_from_geocenter_matches_time_delay_geocentric(self):
        geocenter = np.zeros(3)
        for ifo in self.ifos:
            detector = get_empty_interferometer(ifo).vertex
            for point in np.random.uniform(0, np.pi / 2, (1000, 3)):
                geocentric = geometry.time_delay_geocentric(detector, geocenter, *point)
                from_geocenter = geometry.time_delay_from_geocenter(detector, *point)
                self.assertEqual(from_geocenter, geocentric)

    def test_get_polarization_tensor(self):
        max_diff = 0
        for ra, dec, time, psi in np.random.uniform(0, np.pi / 2, (100, 4)):
            for mode in self.modes:
                args = (ra, dec, time, psi, mode)
                numpy_tensor = get_polarization_tensor(*args)
                cython_tensor = geometry.get_polarization_tensor(*args)
                max_diff = abs(np.max(numpy_tensor - cython_tensor))
        self.assertLess(max_diff, 1e-8)

    def test_get_polarization_tensor_multiple_modes(self):
        max_diff = 0
        for ra, dec, time, psi in np.random.uniform(0, np.pi / 2, (100, 4)):
            args = (ra, dec, time, psi, self.modes)
            numpy_tensors = get_polarization_tensor_multiple_modes(*args)
            cython_tensors = geometry.get_polarization_tensor_multiple_modes(*args)
            max_diff = abs(np.max(np.array(numpy_tensors) - np.array(cython_tensors)))
        self.assertLess(max_diff, 1e-8)

    def test_polarization_tensor_bad_mode_raises_error(self):
        with self.assertRaises(ValueError):
            _ = geometry.get_polarization_tensor(0.0, 0.0, 0.0, 0.0, "bad_mode")

    def test_antenna_reponse(self):
        def _cython_antenna_response(detector_tensor, *args):
            polarization = geometry.get_polarization_tensor(*args)
            return geometry.three_by_three_matrix_contraction(detector_tensor, polarization)

        max_diff = 0
        for ifo in self.ifos:
            detector = get_empty_interferometer(ifo).geometry.detector_tensor
            for ra, dec, time, psi in np.random.uniform(0, np.pi / 2, (100, 4)):
                for mode in self.modes:
                    args = (ra, dec, time, psi, mode)
                    numpy_tensor = antenna_response(detector, *args)
                    cython_tensor = _cython_antenna_response(detector, *args)
                    max_diff = abs(np.max(numpy_tensor - cython_tensor))
        self.assertLess(max_diff, 1e-7)

    def test_frame_conversion(self):
        max_diff = 0
        for ifo_pair in itertools.product(self.ifos, repeat=2):
            if ifo_pair[0] == ifo_pair[1]:
                continue
            ifos = InterferometerList(ifo_pair)
            delta_x = ifos[0].vertex - ifos[1].vertex
            # there is a bug in the python implementation for separations
            # lying in specific quadrants for arctan2 due to using arctan
            # the direct tests of the rotation matrix are more rigorous
            sign_incorrect = abs(np.arctan2(delta_x[1], delta_x[0])) < np.pi / 2
            delta_x *= (-1)**sign_incorrect
            for point in np.random.uniform(0, np.pi / 2, (100, 2)):
                numpy_result = zenith_azimuth_to_theta_phi(*point, delta_x)
                cython_result = geometry.zenith_azimuth_to_theta_phi(*point, delta_x)
                max_diff = max(max_diff, np.max(abs(np.array(numpy_result) - np.array(cython_result))))
        self.assertLess(max_diff, 1e-6)

    def test_calculate_arm(self):
        max_diff = 0
        for point in np.random.uniform(0, 2 * np.pi, (1000, 4)):
            print(geometry.calculate_arm(*point), calculate_arm(*point))
            max_diff = max(max_diff, np.linalg.norm(
                geometry.calculate_arm(*point) - calculate_arm(*point)
            ))
        self.assertLess(max_diff, 1e-10)

    def test_detector_tensor(self):
        for xx, yy in np.random.uniform(0, 1, (1000, 2, 3)):
            numpy_tensor = 0.5 * (np.einsum('i,j->ij', xx, xx) - np.einsum('i,j->ij', yy, yy))
            cython_tensor = geometry.detector_tensor(xx, yy)
            self.assertAlmostEqual(np.min(numpy_tensor - cython_tensor), 0, 6)

    def test_rotation_matrix_transpose_is_inverse(self):
        for delta_x in np.random.uniform(0, 1, (100, 3)):
            rotation = geometry.rotation_matrix_from_delta(delta_x)
            self.assertTrue(np.allclose(rotation.T @ rotation, np.eye(3)))

    def test_rotation_matrix_maps_delta_x_to_z_axis(self):
        for delta_x in np.random.uniform(0, 1, (100, 3)):
            rotation = geometry.rotation_matrix_from_delta(delta_x)
            delta_x /= np.linalg.norm(delta_x)
            self.assertTrue(np.allclose(rotation.T @ delta_x, np.array([0, 0, 1])))
